import argparse
import logging

from .db_to_xml import db2xml
from .xml_to_db import xml2db
from .xml_to_sql import xml2sql


def cli_entry_point():
    """Entry point of the cli application"""

    # Initialize parser for the command line
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command", help="What the cli should do", required=True)

    # Initialize parser for the command to convert db to xml
    parser_db2xml = subparsers.add_parser("db2xml", help="From the db to the xml")
    parser_db2xml.add_argument("db", help="db filename")
    parser_db2xml.add_argument("xml", help="xml filename")
    parser_db2xml.add_argument("-d", "--debug", help="enable debug", action="store_true", required=False)
    parser_db2xml.add_argument(
        "-p",
        "--parse-source-attributes",
        help="enable parsing of source attributes",
        action="store_true",
        required=False,
    )

    # Initialize parser for the command to convert xml to db
    parser_xml2db = subparsers.add_parser("xml2db", help="From the xml to the db.")
    parser_xml2db.add_argument("xml", help="xml filename")
    parser_xml2db.add_argument("db", help="db filename")
    parser_xml2db.add_argument("-d", "--debug", help="enable debug", action="store_true", required=False)
    parser_xml2db.add_argument(
        "-p",
        "--parse-source-attributes",
        help="enable parsing of source attributes",
        action="store_true",
        required=False,
    )
    parser_xml2db.add_argument(
        "-u",
        "--unsafe",
        help="enable the unsafe database insertions, use it carefully to gain speed",
        action="store_true",
        required=False,
    )

    # Initialize parser for the command to convert xml to sql
    parser_xml2sql = subparsers.add_parser("xml2sql", help="From the xml to the sql.")
    parser_xml2sql.add_argument("xml", help="xml filename")
    parser_xml2sql.add_argument("sql", help="sql filename")
    parser_xml2sql.add_argument("-d", "--debug", help="enable debug", action="store_true", required=False)
    parser_xml2sql.add_argument(
        "-p",
        "--parse-source-attributes",
        help="enable parsing of source attributes",
        action="store_true",
        required=False,
    )

    # Initialize parser for the command to test the application
    parser_test = subparsers.add_parser("test", help="Test if the console program is installed.")
    parser_test.add_argument("-d", "--debug", help="enable debug", action="store_true", required=False)

    # Parse arguments
    args = parser.parse_args()

    if args.debug:
        # Start the logging via terminal
        logging.basicConfig(level=logging.DEBUG, format="%(message)s")

    # Execute the command given in the arguments
    if args.command == "db2xml":
        cli_db2xml(db=args.db, xml=args.xml, p=args.parse_source_attributes, debug=args.debug)

    elif args.command == "xml2db":
        cli_xml2db(db=args.db, xml=args.xml, p=args.parse_source_attributes, unsafe=args.unsafe, debug=args.debug)

    elif args.command == "xml2sql":
        cli_xml2sql(sql=args.sql, xml=args.xml, p=args.parse_source_attributes, debug=args.debug)

    elif args.command == "test":
        print("sqlerddap installed correctly.")


def check_xml_extension(xml: str) -> bool:
    """Check if the given xml filename has the right extension and print some debug info.

    Args:
        xml (str): xml filename to check.

    Returns:
        bool: the given xml filename is valid
    """

    valid = xml.endswith(".xml")
    if not valid:
        logging.debug("ATTENTION: the given xml file doesn't match the xml extension.")
    return valid


def check_db_extension(db: str) -> bool:
    """Check if the given db filename has the right extension and print some debug info.

    Args:
        db (str): db filename to check.

    Returns:
        bool: the given db filename is valid
    """

    valid = db.endswith(".db") or db.endswith(".sqlite") or db.endswith(".db3")
    if not (valid):
        logging.debug("ATTENTION: the given db file doesn't match the db extension: [.sqlite, .db, .db3]")
    return valid


def check_sql_extension(sql: str) -> bool:
    """Check if the given sql filename has the right extension and print some debug info.

    Args:
        sql (str): sql filename to check.

    Returns:
        bool: the given sql filename is valid
    """

    valid = sql.endswith(".sql")
    if not (valid):
        logging.debug("ATTENTION: the given sql filename doesn't match the sql extension: .sql")
    return valid


def cli_db2xml(db: str, xml: str, p: bool, debug: bool) -> None:
    """Execute the db to xml command with additional cli procedures.

    Args:
        db (str): database filename.
        xml (str): xml filename.
        p (bool): parse_source_attributes argument.
        debug (bool): debug argument.
    """

    if debug:
        check_xml_extension(xml)
        check_db_extension(db)

    db2xml(db_filename=db, xml_filename=xml, parse_source_attributes=p)


def cli_xml2db(db: str, xml: str, p: bool, unsafe: bool, debug: bool) -> None:
    """Execute the xml to db command with additional cli procedures.

    Args:
        db (str): database filename.
        xml (str): xml filename.
        p (bool): parse_source_attributes argument.
        debug (bool): debug argument.
    """

    if debug:
        check_xml_extension(xml)
        check_db_extension(db)

    xml2db(db_filename=db, xml_filename=xml, parse_source_attributes=p, unsafe=unsafe)


def cli_xml2sql(sql: str, xml: str, p: bool, debug: bool) -> None:
    """Execute the xml to sql command with additional cli procedures.

    Args:
        db (str): database filename.
        xml (str): xml filename.
        p (bool): parse_source_attributes argument.
        debug (bool): debug argument.
    """

    if debug:
        check_xml_extension(xml)
        check_sql_extension(sql)

    xml2sql(sql_filename=sql, xml_filename=xml, parse_source_attributes=p)



from .xml_to_obj import xml2obj
from .obj_to_db import obj2db


def xml2db(db_filename: str, xml_filename: str, parse_source_attributes: bool = False, unsafe: bool = False) -> None:
    """Convert a XML datasets ERDDAP configuration to a DB sqlite.

    Args:
        db_filename (str): database sqlite filename.
        xml_filename (str): xml filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
        unsafe (bool, optional): Flag to enable unsafe execution to the database, this will disable journal and synchronous. Defaults to False.
    """

    erddap = xml2obj(
        xml_filename=xml_filename,
        parse_source_attributes=parse_source_attributes,
    )

    obj2db(
        erddap=erddap,
        db_filename=db_filename,
        parse_source_attributes=parse_source_attributes,
        unsafe=unsafe,
    )



from enum import StrEnum


class Mode(StrEnum):
    """Enum used to select the mode of the line ending conversion."""

    Win2Unix = "Win2Unix"
    Unix2Win = "Unix2Win"


def change_line_ending(filename: str, mode: Mode = Mode.Win2Unix):
    """Convert the line ending of the given file using the given mode.

    Args:
        filename (str): File to convert the line ending.
        mode (Mode, optional): Type of conversion, windows to unix or vice versa. Defaults to Mode.Win2Unix.
    """
    WINDOWS_LINE_ENDING = b"\r\n"
    UNIX_LINE_ENDING = b"\n"

    # Open given file
    with open(filename, "rb") as open_file:
        content = open_file.read()

    if mode == Mode.Win2Unix:
        # Windows to Unix
        content = content.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

    elif mode == Mode.Unix2Win:
        # Unix to Windows
        content = content.replace(UNIX_LINE_ENDING, WINDOWS_LINE_ENDING)

    # Save converted file
    with open(filename, "wb") as open_file:
        open_file.write(content)



"""This module contains the some sql scripts."""

SQL_UNSAFE = """
PRAGMA journal_mode = OFF;
PRAGMA synchronous = 0;
-- The followings has to be manage in some way and tested
-- PRAGMA cache_size = 1000000;
-- PRAGMA locking_mode = EXCLUSIVE;
-- PRAGMA temp_store = MEMORY;
"""

SQL_CREATE = """
CREATE TABLE IF NOT EXISTS params (
    param_name TEXT PRIMARY KEY NOT NULL,
    param_value TEXT DEFAULT ''
);

CREATE TABLE IF NOT EXISTS datasets (
    dataset_id TEXT PRIMARY KEY NOT NULL,
    dataset_type TEXT NOT NULL,
    dataset_active TEXT CHECK(dataset_active='true' or dataset_active='false') NOT NULL DEFAULT 'false'
);

CREATE TABLE IF NOT EXISTS dataset_children (
    parent_dataset_id TEXT NOT NULL,
    child_dataset_id TEXT NOT NULL,
    FOREIGN KEY(parent_dataset_id) REFERENCES datasets(dataset_id),
    FOREIGN KEY(child_dataset_id) REFERENCES datasets(dataset_id)
);

CREATE TABLE IF NOT EXISTS dataset_params (
    param_name TEXT NOT NULL,
    param_value TEXT DEFAULT '',
    dataset_id TEXT NOT NULL,
    PRIMARY KEY (param_name, dataset_id),
    FOREIGN KEY(dataset_id) REFERENCES datasets(dataset_id)
);

CREATE TABLE IF NOT EXISTS dataset_attributes (
    attribute_name TEXT NOT NULL,
    attribute_type TEXT DEFAULT '',
    attribute_value TEXT DEFAULT '',
    dataset_id TEXT NOT NULL,
    PRIMARY KEY (attribute_name, dataset_id),
    FOREIGN KEY(dataset_id) REFERENCES datasets(dataset_id)
);

CREATE TABLE IF NOT EXISTS dataset_source_attributes (
    attribute_name TEXT NOT NULL,
    attribute_type TEXT DEFAULT '',
    attribute_value TEXT DEFAULT '',
    dataset_id TEXT NOT NULL,
    PRIMARY KEY (attribute_name, dataset_id),
    FOREIGN KEY(dataset_id) REFERENCES datasets(dataset_id)
);

CREATE TABLE IF NOT EXISTS variables (
    variable_id INTEGER PRIMARY KEY NOT NULL,
    category TEXT CHECK(category='data' or category='axis') DEFAULT 'data',
    source_name TEXT NOT NULL,
    destination_name TEXT NOT NULL,
    data_type TEXT NOT NULL,
    dataset_id TEXT NOT NULL,
    order_number REAL NOT NULL DEFAULT 0,
    FOREIGN KEY(dataset_id) REFERENCES datasets(dataset_id)
);

CREATE TABLE IF NOT EXISTS variable_attributes (
    attribute_name TEXT NOT NULL,
    attribute_type TEXT DEFAULT '',
    attribute_value TEXT DEFAULT '',
    variable_id INTEGER NOT NULL,
    PRIMARY KEY (attribute_name, variable_id),
    FOREIGN KEY(variable_id) REFERENCES variables(variable_id)
);

CREATE TABLE IF NOT EXISTS variable_source_attributes (
    attribute_name TEXT NOT NULL,
    attribute_type TEXT DEFAULT '',
    attribute_value TEXT DEFAULT '',
    variable_id INTEGER NOT NULL,
    PRIMARY KEY (attribute_name, variable_id),
    FOREIGN KEY(variable_id) REFERENCES variables(variable_id)
);
"""



from dataclasses import dataclass


@dataclass
class PARAM:
    name: str
    value: str


@dataclass
class ATTRIBUTE:
    name: str
    text: str
    type: str = ""


@dataclass
class VARIABLE:
    tag: str
    sourceName: str
    destinationName: str
    dataType: str
    attributes: list[ATTRIBUTE] | None
    source_attributes: list[ATTRIBUTE] | None


@dataclass
class DATASET:
    type: str
    datasetID: str
    active: str
    datasets: list | None
    params: list[PARAM] | None
    attributes: list[ATTRIBUTE] | None
    source_attributes: list[ATTRIBUTE] | None
    variables: list[VARIABLE] | None


@dataclass
class ERDDAP:
    params: list[PARAM]
    datasets: list[DATASET]
    parent_child: list[tuple[2]]


def print_dataset(dataset: DATASET, indent: int):
    indent_str = " ".join(["" for i in range(indent)])
    print()
    print("dataset:")
    print(f"{indent_str} datasetID: {dataset.datasetID}")
    print(f"{indent_str} type: {dataset.type}")
    print(f"{indent_str} active: {dataset.active}")
    print(f"{indent_str} datasets:")
    for d in dataset.datasets:
        print_dataset(d, indent + 2)
    print(f"{indent_str} params:")
    for p in dataset.params:
        print(f"{indent_str}  {p}")
    print(f"{indent_str} attributes:")
    for a in dataset.attributes:
        print(f"{indent_str}  {a}")
    print(f"{indent_str} variables:")
    for v in dataset.variables:
        print(f"{indent_str}  {v.tag} {v.dataType} {v.destinationName} {v.sourceName}")
        print(f"{indent_str}   attributes:")
        for a in dataset.attributes:
            print(f"{indent_str}    {a}")



from .classes import ERDDAP
from .sql_script import SQL_CREATE
from .template_utils import obj2sql_string


def obj2sql(erddap: ERDDAP, sql_filename: str, parse_source_attributes: bool = False) -> None:
    """Convert a XML datasets ERDDAP configuration to a sql script.

    Args:
        erddap (ERDDAP): python object to convert.
        sql_filename (str): sql filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    output = obj2sql_string(erddap)

    # Save sql
    with open(sql_filename, "w", encoding="utf-8") as f:
        f.write(SQL_CREATE)
        f.write(output)



from jinja2 import Environment, PackageLoader, select_autoescape
from .classes import ERDDAP


def obj2sql_string(erddap: ERDDAP):
    """Convert an ERDDAP object into a sql script INSERT script, using the jinja2 template engine.

    Args:
        erddap (ERDDAP): ERDDAP object to convert into a sql script.
    """
    # Render the template
    env = Environment(loader=PackageLoader("erddapcfg"), autoescape=select_autoescape())
    template = env.get_template("db_insert.j2")
    output = template.render(
        params=erddap.params,
        datasets=erddap.datasets,
        dataset_params=[
            {"name": param.name, "value": param.value, "datasetID": dataset.datasetID}
            for dataset in erddap.datasets
            for param in dataset.params
        ],
        dataset_attributes=[
            {
                "name": attribute.name,
                "type": attribute.type,
                "text": attribute.text,
                "datasetID": dataset.datasetID,
            }
            for dataset in erddap.datasets
            for attribute in dataset.attributes
        ],
        dataset_source_attributes=[
            {
                "name": attribute.name,
                "type": attribute.type,
                "text": attribute.text,
                "datasetID": dataset.datasetID,
            }
            for dataset in erddap.datasets
            for attribute in dataset.source_attributes
        ],
        variables=[
            {
                "tag": variable.tag,
                "sourceName": variable.sourceName,
                "destinationName": variable.destinationName,
                "dataType": variable.dataType,
                "datasetID": dataset.datasetID,
            }
            for dataset in erddap.datasets
            for variable in dataset.variables
        ],
        variable_attributes=[
            {
                "name": attribute.name,
                "type": attribute.type,
                "text": attribute.text,
                "destinationName": variable.destinationName,
                "datasetID": dataset.datasetID,
            }
            for dataset in erddap.datasets
            for variable in dataset.variables
            for attribute in variable.attributes
        ],
        variable_source_attributes=[
            {
                "name": attribute.name,
                "type": attribute.type,
                "text": attribute.text,
                "destinationName": variable.destinationName,
                "datasetID": dataset.datasetID,
            }
            for dataset in erddap.datasets
            for variable in dataset.variables
            for attribute in variable.source_attributes
        ],
        dataset_children=erddap.parent_child,
    )

    return output


def obj2xml_string(erddap: ERDDAP):
    """Convert an ERDDAP object into a xml string, using the jinja2 template engine.

    Args:
        erddap (ERDDAP): ERDDAP object to convert into a xml string.
    """
    env = Environment(loader=PackageLoader("erddapcfg"), autoescape=select_autoescape())
    template = env.get_template("datasets.xml.j2")
    output = "".join(["<?xml version='1.0' encoding='ISO-8859-1'?>\n", template.render(erddap=erddap)])
    return output



import os
import sqlite3

from .classes import ERDDAP
from .sql_script import SQL_UNSAFE, SQL_CREATE
from .template_utils import obj2sql_string


def obj2db(erddap: ERDDAP, db_filename: str, parse_source_attributes: bool = False, unsafe: bool = False) -> None:
    """Convert a XML datasets ERDDAP configuration to a DB sqlite.

    Args:
        erddap (ERDDAP): python object to convert.
        db_filename (str): database sqlite filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
        unsafe (bool, optional): Flag to enable unsafe execution to the database, this will disable journal and synchronous. Defaults to False.
    """

    script = [
        SQL_UNSAFE if unsafe else "",
        SQL_CREATE,
        obj2sql_string(erddap),
    ]

    # Create empty database file
    if os.path.isfile(db_filename):
        os.remove(db_filename)
    connection = sqlite3.connect(db_filename)
    cursor = connection.cursor()

    # Execute sql script
    cursor.executescript("".join(script))

    # Save db
    connection.commit()
    connection.close()



from .xml_to_obj import xml2obj
from .obj_to_sql import obj2sql


def xml2sql(sql_filename: str, xml_filename: str, parse_source_attributes: bool = False) -> None:
    """Convert a XML datasets ERDDAP configuration to a sql script.

    Args:
        sql_filename (str): sql filename.
        xml_filename (str): xml filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    erddap = xml2obj(xml_filename=xml_filename, parse_source_attributes=parse_source_attributes)

    obj2sql(erddap=erddap, sql_filename=sql_filename, parse_source_attributes=parse_source_attributes)



from setuptools import setup

setup(
    name="erddapcfg",
    version="0.0.2",
    description="Python program to work with ERDDAP configurations.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/PlehanSebastian/erddapcfg",
    author="Sebastian Plehan",
    author_email="plehan.sebastian@gmail.com",
    packages=[
        "erddapcfg",
    ],
    package_dir={"erddapcfg": "erddapcfg"},
    entry_points={
        "console_scripts": ["erddapcfg=erddapcfg.cli:cli_entry_point"],
    },
    include_package_data=True,
    install_requires=[
        "pandas",
        "jinja2",
    ],
    data_files=[
        (
            "",
            [
                "erddapcfg/templates/datasets.xml.j2",
                "erddapcfg/templates/db_insert.j2",
                "erddapcfg/templates/macro.j2",
            ],
        )
    ],
)



from xml.sax.saxutils import escape
import sqlite3
import logging

import pandas as pd

from .classes import ERDDAP, PARAM, DATASET, ATTRIBUTE, VARIABLE


def db2obj(db_filename: str, parse_source_attributes: bool = False) -> ERDDAP:
    """Convert a DB sqlite to a python object.

    Args:
        db_filename (str): database sqlite filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    # Initialize objects for jinja template engine
    erddap = ERDDAP(
        params=[],
        datasets=[],
        parent_child=[],
    )

    # Open database
    connection = sqlite3.connect(db_filename)

    # Loading dataframes from database
    df_params = pd.read_sql_query(
        sql="SELECT * FROM params",
        con=connection,
    )
    df_datasets = pd.read_sql_query(
        sql="SELECT * FROM datasets",
        con=connection,
    )
    df_dataset_params = pd.read_sql_query(
        sql="SELECT * FROM dataset_params",
        con=connection,
    )
    df_dataset_attributes = pd.read_sql_query(
        sql="SELECT * FROM dataset_attributes",
        con=connection,
    )
    df_dataset_source_attributes = pd.read_sql_query(
        sql="SELECT * FROM dataset_source_attributes",
        con=connection,
    )
    df_variables = pd.read_sql_query(
        sql="SELECT * FROM variables",
        con=connection,
    )
    df_variable_attributes = pd.read_sql_query(
        sql="SELECT * FROM variable_attributes JOIN variables USING (variable_id)",
        con=connection,
    )
    df_variable_source_attributes = pd.read_sql_query(
        sql="SELECT * FROM variable_source_attributes JOIN variables USING (variable_id)",
        con=connection,
    )
    df_dataset_children = pd.read_sql_query(
        sql="SELECT dataset_id, child_dataset_id FROM datasets JOIN dataset_children ON datasets.dataset_id==dataset_children.parent_dataset_id",
        con=connection,
    )

    # Load params from db to objects
    for _, row in df_params.iterrows():
        erddap.params.append(
            PARAM(
                name=row["param_name"],
                value=row["param_value"] if row["param_value"] is not None else "",
            )
        )

    # Load datasets from db to objects
    for _, row in df_datasets.iterrows():
        erddap.datasets.append(
            DATASET(
                type=row["dataset_type"],
                datasetID=row["dataset_id"],
                active=row["dataset_active"],
                datasets=[],
                params=[],
                attributes=[],
                source_attributes=[],
                variables=[],
            )
        )

    # Load params, attributes, variables for each dataset from db to objects
    for dataset in erddap.datasets:
        logging.debug(dataset.datasetID)

        # Load dataset params from db to objects
        df = df_dataset_params[df_dataset_params["dataset_id"] == dataset.datasetID]
        for _, row in df.iterrows():
            dataset.params.append(
                PARAM(
                    name=row["param_name"],
                    value=row["param_value"] if row["param_value"] is not None else "",
                )
            )

        # Load dataset attributes from db to objects
        df = df_dataset_attributes[df_dataset_attributes["dataset_id"] == dataset.datasetID]
        for _, row in df.iterrows():
            dataset.attributes.append(
                ATTRIBUTE(
                    name=row["attribute_name"],
                    text=row["attribute_value"] if row["attribute_value"] is not None else "",
                    type=row["attribute_type"],
                )
            )

        # Load dataset source attributes from db to objects
        if parse_source_attributes:
            df = df_dataset_source_attributes[df_dataset_source_attributes["dataset_id"] == dataset.datasetID]
            for _, row in df.iterrows():
                dataset.source_attributes.append(
                    ATTRIBUTE(
                        name=row["attribute_name"],
                        text=row["attribute_value"] if row["attribute_value"] is not None else "",
                        type=row["attribute_type"],
                    )
                )

        # Load variables from db to objects
        df = df_variables[df_variables["dataset_id"] == dataset.datasetID]
        for _, row in df.iterrows():
            dataset.variables.append(
                VARIABLE(
                    tag=row["category"] + "Variable",
                    sourceName=row["source_name"],
                    destinationName=row["destination_name"],
                    dataType=row["data_type"],
                    attributes=None,
                    source_attributes=None,
                )
            )

        # Load attributes and source attributes for each variable from db to objects
        for variable in dataset.variables:
            # Load variable attributes from db to objects
            df = df_variable_attributes[df_variable_attributes["dataset_id"] == dataset.datasetID]
            df = df[df["destination_name"] == variable.destinationName]
            for _, row in df.iterrows():
                if variable.attributes is None:
                    variable.attributes = []

                variable.attributes.append(
                    ATTRIBUTE(
                        name=row["attribute_name"],
                        text=escape(str(row["attribute_value"]).replace('"', ""))
                        if row["attribute_value"] is not None
                        else "",
                        type=row["attribute_type"],
                    )
                )

            # Load variable source attributes from db to objects
            if parse_source_attributes:
                df = df_variable_source_attributes[df_variable_source_attributes["dataset_id"] == dataset.datasetID]
                df = df[df["destination_name"] == variable.destinationName]
                for _, row in df.iterrows():
                    if variable.source_attributes is None:
                        variable.source_attributes = []

                    variable.source_attributes.append(
                        ATTRIBUTE(
                            name=row["attribute_name"],
                            text=escape(str(row["attribute_value"]).replace('"', ""))
                            if row["attribute_value"] is not None
                            else "",
                            type=row["attribute_type"],
                        )
                    )

    # Load the parent - child relations between datasets from db to objects
    for _, row in df_dataset_children.iterrows():
        dataset_parent = [d for d in erddap.datasets if d.datasetID == row["dataset_id"]][0]
        dataset_child = [d for d in erddap.datasets if d.datasetID == row["child_dataset_id"]][0]
        dataset_parent.datasets.append(dataset_child)
        erddap.datasets.remove(dataset_child)

    return erddap



from .db_to_obj import db2obj
from .obj_to_xml import obj2xml


def db2xml(db_filename: str, xml_filename: str, parse_source_attributes: bool = False) -> None:
    """Convert a DB sqlite to a XML datasets ERDDAP configuration.

    Args:
        db_filename (str): database sqlite filename.
        xml_filename (str): xml filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    erddap = db2obj(db_filename=db_filename, parse_source_attributes=parse_source_attributes)

    obj2xml(erddap=erddap, xml_filename=xml_filename, parse_source_Attributes=parse_source_attributes)



from .classes import ERDDAP
from .utils import change_line_ending
from .template_utils import obj2xml_string


def obj2xml(erddap: ERDDAP, xml_filename: str, parse_source_Attributes: bool = False) -> None:
    """Convert a python object to a XML datasets ERDDAP configuration.

    Args:
        erddap (ERDDAP): python object to convert.
        xml_filename (str): xml filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    # Render the template
    output = obj2xml_string(erddap=erddap)

    # Custom unescape the CDATA blocks
    output = output.replace("::CDATA_START", "<![CDATA[")
    output = output.replace("::CDATA_STOP", "]]>")

    # Save xml
    with open(xml_filename, "w", encoding="utf-8") as f:
        f.write(output)
    change_line_ending(xml_filename)



from .db_to_obj import db2obj
from .db_to_xml import db2xml

from .xml_to_obj import xml2obj
from .xml_to_db import xml2db
from .xml_to_sql import xml2sql

from .obj_to_xml import obj2xml
from .obj_to_db import obj2db
from .obj_to_sql import obj2sql



import xml.etree.ElementTree as ET
import logging

from .classes import ERDDAP, PARAM, DATASET, ATTRIBUTE, VARIABLE


def xml2obj(xml_filename: str, parse_source_attributes: bool = False) -> ERDDAP:
    """Convert a XML datasets ERDDAP configuration to python objects.

    Args:
        xml_filename (str): xml filename.
        parse_source_Attributes (bool, optional): Flag to enable the parsing of the sourceAttributes nodes. Defaults to False.
    """

    # Initialize erddap object
    erddap = ERDDAP(
        params=[],
        datasets=[],
        parent_child=[],
    )

    # Load XML file
    with open(xml_filename, "r") as file:
        xml_string = "".join(file.readlines())

        # Custom escape CDATA blocks
        xml_string = xml_string.replace("<![CDATA[", "<![CDATA[::CDATA_START").replace("]]>", "::CDATA_STOP]]>")

        # Uncomment sourceAttributes blocks to make it readable from the parser
        if parse_source_attributes:
            xml_string = xml_string.replace("<!-- sourceAttributes>", "<sourceAttributes>").replace(
                "</sourceAttributes -->", "</sourceAttributes>"
            )

    datasets_root: ET.Element = ET.XML(xml_string)

    # Load params from xml to objects
    for child in datasets_root:
        if child.tag == "dataset":
            continue

        erddap.params.append(
            PARAM(
                name=child.tag,
                value=child.text if child.text is not None else "",
            )
        )

    # Load datasets from xml to objects
    children = datasets_root.findall(".//dataset")
    for child in children:
        erddap.datasets.append(
            DATASET(
                type=child.attrib["type"],
                datasetID=child.attrib["datasetID"],
                active=child.attrib["active"] if "active" in child.attrib else "true",
                datasets=[],
                params=[],
                attributes=[],
                source_attributes=[],
                variables=[],
            )
        )

    # Load params, attributes, variables for each dataset from xml to objects
    for dataset in erddap.datasets:
        logging.debug(dataset.datasetID)
        dataset_node = datasets_root.find(f".//dataset[@datasetID='{dataset.datasetID}']")

        # Load dataset params from xml to objects
        params_nodes = [
            child
            for child in dataset_node
            if child.tag not in ("dataset", "addAttributes", "sourceAttributes", "dataVariable", "axisVariable")
        ]
        for node in params_nodes:
            dataset.params.append(
                PARAM(
                    name=node.tag,
                    value=node.text if node.text is not None else "",
                )
            )

        # Load dataset attributes from xml to objects
        attribute_node = dataset_node.find("addAttributes")
        if attribute_node is not None:
            for node in attribute_node:
                dataset.attributes.append(
                    ATTRIBUTE(
                        name=node.attrib["name"],
                        text=node.text if node.text is not None else "",
                        type=node.attrib["type"] if "type" in node.attrib else "",
                    )
                )

        # Load dataset source attributes from xml to objects
        if parse_source_attributes:
            source_attribute_node = dataset_node.find("sourceAttributes")
            if source_attribute_node is not None:
                for node in source_attribute_node:
                    dataset.source_attributes.append(
                        ATTRIBUTE(
                            name=node.attrib["name"],
                            text=node.text if node.text is not None else "",
                            type=node.attrib["type"] if "type" in node.attrib else "",
                        )
                    )

        # Load dataset variables from xml to objects
        variable_nodes = dataset_node.findall("dataVariable") + dataset_node.findall("axisVariable")
        if variable_nodes is not None:
            for node in variable_nodes:
                source_name_node = node.find("sourceName")
                destination_name_node = node.find("destinationName")
                data_type_node = node.find("dataType")
                dataset.variables.append(
                    VARIABLE(
                        tag=node.tag,
                        sourceName=source_name_node.text if source_name_node is not None else "",
                        destinationName=destination_name_node.text if destination_name_node is not None else "",
                        dataType=data_type_node.text if data_type_node is not None else "",
                        attributes=[],
                        source_attributes=[],
                    )
                )

                # Load variable attributes from xml to objects
                attributes_node = node.find("addAttributes")
                if attributes_node is not None:
                    for child in attributes_node:
                        dataset.variables[-1].attributes.append(
                            ATTRIBUTE(
                                name=child.attrib["name"],
                                text=child.text if child.text is not None else "",
                                type=child.attrib["type"] if "type" in child.attrib else "",
                            )
                        )

                # Load variable source attributes from xml to objects
                if parse_source_attributes:
                    source_attribute_node = node.find("sourceAttributes")
                    if source_attribute_node is not None:
                        for node in source_attribute_node:
                            dataset.source_attributes.append(
                                ATTRIBUTE(
                                    name=node.attrib["name"],
                                    text=node.text if node.text is not None else "",
                                    type=node.attrib["type"] if "type" in node.attrib else "",
                                )
                            )

        # Load the parent - child relations between datasets from xml to objects
        dataset_children = dataset_node.findall("dataset")
        if dataset_children is not None:
            for dataset_child in dataset_children:
                erddap.parent_child.append((dataset.datasetID, dataset_child.attrib["datasetID"]))

    return erddap
